@@ -1,8 +1,4 @@
 import logging
-from pathlib import Path
-from subprocess import Popen
-from subprocess import PIPE
-from subprocess import STDOUT
 import os
 import shutil
 from copy import deepcopy
@@ -14,16 +10,20 @@ import importlib.util
 import types
 import traceback
 import jsonpickle
+from pathlib import Path
+from subprocess import Popen
+from subprocess import PIPE
+from subprocess import STDOUT
 import inspect
+import ctypes
+import operator
+import re
 import argparse
 import requests
 import yaml
 from tqdm import tqdm
 from zipfile import ZipFile
 from distutils.dir_util import mkpath
-import operator
-import ctypes
-import re
 
 ######## START CONTENT ########
 def INVALID_NAME():
@@ -714,12 +714,12 @@ class Functor(Datum):
 			else:
 				logging.warning(f"{this.name} was not given an 'executor'. Some features will not be available.")
 
-			if ('precursor' in this.kwargs):
-				this.precursor = this.kwargs.pop('precursor')
-				logging.debug(f"{this.name} was preceded by {this.precursor.name}")
-			else:
-				this.precursor = None
-				logging.debug(f"{this.name} was preceded by None")
+		if ('precursor' in this.kwargs):
+			this.precursor = this.kwargs.pop('precursor')
+			logging.debug(f"{this.name} was preceded by {this.precursor.name}")
+		else:
+			this.precursor = None
+			logging.debug(f"{this.name} was preceded by None")
 
 
 	# Override this with any additional argument validation you need.
@@ -1244,6 +1244,65 @@ class StandardFunctor(Functor):
 	######## END: UTILITIES ########
 
 
+# ExecutorTracker is a global singleton which keeps a record of all Executors that have been launched.
+# This can be abused quite a bit, so please try to restrict usage of this to only:
+# * Ease of use global functions
+#
+# Thanks! 
+class ExecutorTracker:
+	def __init__(this):
+		# Singletons man...
+		if "instance" not in ExecutorTracker.__dict__:
+			logging.debug(f"Creating new ExecutorTracker: {this}")
+			ExecutorTracker.instance = this
+		else:
+			return None
+
+		this.executors = [None]
+
+	@staticmethod
+	def Instance():
+		if "instance" not in ExecutorTracker.__dict__:
+			ExecutorTracker()
+		return ExecutorTracker.instance
+
+	@staticmethod
+	def Push(executor):
+		ExecutorTracker.Instance().executors.append(executor)
+
+		# Adding the executor to our list here increases its reference count.
+		# Executors are supposed to remove themselves from this list when they are deleted.
+		# A python object cannot be deleted if it has references.
+		# Thus, we forcibly decrease the reference count and rely on Exectuor's self-reporting to avoid accessing deallocated memory.
+		# This appears to cause segfaults on some systems, so we'll just live with the fact that Executors will never be destroyed.
+		# If you want your executor to stop being tracked, do it yourself. :(
+		#
+		# ctypes.pythonapi.Py_DecRef(ctypes.py_object(executor))
+
+		logging.debug(f"Now tracking Executor: {executor}")
+
+	@staticmethod
+	def Pop(executor):
+		try:
+			ExecutorTracker.Instance().executors.remove(executor)
+			logging.debug(f"No longer tracking Executor: {executor}")
+		except:
+			pass
+
+	@staticmethod
+	def GetLatest():
+		return ExecutorTracker.Instance().executors[-1]
+
+
+# Global Fetch() function.
+# Uses the latest registered Executor
+def Fetch(varName, default=None, fetchFrom=None, start=True, attempted=None):
+    return ExecutorTracker.GetLatest().Fetch(varName, default, fetchFrom, start, attempted)
+
+# Ease-of-use wrapper for the global Fetch()
+def f(varName, default=None, fetchFrom=None, start=True, attempted=None):
+    Fetch(varName, default, fetchFrom, start, attempted)
+
 # A DataContainer allows Data to be stored and worked with.
 # This class is intended to be derived from and added to.
 # Each DataContainer is comprised of multiple Data (see Datum.py for more).
@@ -1402,6 +1461,159 @@ class DataContainer(Datum):
 		return []
 
 
+
+# Use an ErrorStringParser for each "parsers" in order to avoid having to override the GetObjectFromError method and create a new class for every error you want to handle.
+# ErrorStringParsers enable ErrorResolutions to be created on a per-functionality, rather than per-error basis, reducing the total amount of duplicate code.
+# Each error has a different string. In order to get the object of the error, we have to know where the object starts and ends.
+# NOTE: this assumes only 1 object per string. Maybe fancier parsing logic can be added in the future.
+#
+# startPosition is always positive
+# endPosition is always negative
+class ErrorStringParser:
+
+	def __init__(this, applicableError, startPosition, endPosition):
+		this.applicableError = applicableError
+		this.startPosition = startPosition
+		this.endPosition = endPosition
+
+	def Parse(this, errorString):
+		end = this.endPosition
+		if (not end):
+			end = len(errorString)
+		return errorString[this.startPosition:end]
+
+
+# ErrorResolution is a Functor which can be executed when an Exception is raised.
+# The goal of this class is to do some kind of work that will fix the problem on the second try of whatever generated the error.
+class ErrorResolution(StandardFunctor):
+
+	def __init__(this, name=INVALID_NAME()):
+		super().__init__(name)
+
+		# What errors, as ErrorStringParser objects, is *this prepared to handle?
+		this.parsers = []
+
+		this.error = None
+		this.errorType = ""
+		this.errorString = ""
+		this.errorObject = ""
+		this.errorResolutionStack = {}
+
+		# Provided directly from the recoverable decorator.
+		this.optionalKWArgs["obj"] = None
+		this.optionalKWArgs["function"] = None
+
+		# We do want to know whether or not we should attempt to run whatever failed again.
+		# So, let's store that in functionSucceeded. Meaning if this.functionSucceeded, try the original method again.
+		# No rollback, by default and definitely don't throw Exceptions.
+		this.enableRollback = False
+		this.functionSucceeded = True
+		this.raiseExceptions = False
+
+		this.errorShouldBeResolved = False
+
+
+
+	# Put your logic here!
+	def Resolve(this):
+		# You get the following members:
+		# this.error (an Exception)
+		# this.errorString (a string cast of the Exception)
+		# this.errorType (a string)
+		# this.errorObjet (a string or whatever you return from GetObjectFromError())
+
+		# You get the following guarantees:
+		# *this has not been called on this particular error before.
+		# the error given is applicable to *this per this.parsers
+
+		###############################################
+		# Please throw errors if something goes wrong #
+		# Otherwise, set this.errorShouldBeResolved   #
+		###############################################
+		
+		pass
+
+
+
+	# Helper method for creating ErrorStringParsers
+	# To use this, simply take an example output and replace the object you want to extract with "OBJECT"
+	def ApplyTo(this, error, exampleString):
+		match = re.search('OBJECT', exampleString)
+		this.parsers.append(ErrorStringParser(error, match.start(), match.end() - len(exampleString)))
+
+
+	# Get the type of this.error as a string.
+	def GetErrorType(this):
+		return type(this.error).__name__
+
+
+	# Get an actionable object from the error.
+	# For example, if the error is 'ModuleNotFoundError', what is the module?
+	def GetObjectFromError(this):
+		for parser in this.parsers:
+			if (parser.applicableError != this.errorType):
+				continue
+
+			this.errorObject = parser.Parse(this.errorString)
+			return
+
+		raise ErrorResolutionError(f"{this.name} cannot parse error object from ({this.errorType}): {str(this.error)}.")
+
+
+	# Determine if this resolution method is applicable.
+	def CanProcess(this):
+		return this.GetErrorType() in [parser.applicableError for parser in this.parsers]
+
+
+	# Grab any known and necessary args from this.kwargs before any Fetch calls are made.
+	def ParseInitialArgs(this):
+		super().ParseInitialArgs()
+		if ('error' in this.kwargs):
+			this.error = this.kwargs.pop('error')
+			# Just assume the error is an actual Exception object.
+		else:
+			raise ErrorResolutionError(f"{this.name} was not given an error to resolve.")
+
+		this.errorString = str(this.error)
+		this.errorType = this.GetErrorType()
+
+		# Internal member to avoid processing duplicates
+		this.errorResolutionStack = this.executor.errorResolutionStack
+
+
+	# Error resolution is unchained.
+	def PopulateNext(this):
+		this.next = []
+
+
+	# Override of Functor method.
+	# We'll keep calling this until an error is raised.
+	def Function(this):
+		this.functionSucceeded = True
+		this.errorShouldBeResolved = True
+		
+		if (not this.CanProcess()):
+			this.errorShouldBeResolved = False
+			return this.errorResolutionStack, this.errorShouldBeResolved
+
+		if (not this.errorString in this.errorResolutionStack.keys()):
+			this.errorResolutionStack.update({this.errorString:[]})
+		
+		if (this.name in this.errorResolutionStack[this.errorString]):
+			raise FailedErrorResolution(f"{this.name} already tried and failed to resolve {this.errorType}: {this.errorString}.")
+
+		this.GetObjectFromError()
+
+		try:
+			this.Resolve()
+		except Exception as e:
+			logging.error(f"Error resolution with {this.name} failed: {e}")
+			util.LogStack()
+			this.functionSucceeded = False
+		
+		this.errorResolutionStack[this.errorString].append(this.name)
+		return this.errorResolutionStack, this.errorShouldBeResolved
+
 #from .Executor import Executor # don't import this, it'll be circular!
 
 # @recoverable
@@ -1423,95 +1635,63 @@ def RecoverableImplementation(obj, executor, function, *args, **kwargs):
 	except Exception as e:
 		if (not executor.resolveErrors):
 			raise e
+		return Recover(e, obj, executor, function, *args, **kwargs)
 
-		logging.warning(f"Got error '{e}' from function ({function}) by {obj.name}.")
-		util.LogStack()
 
-		# We have to use str(e) instead of pointers to Exception objects because multiple Exceptions will have unique addresses but will still be for the same error, as defined by string comparison.
-		if (str(e) not in executor.errorResolutionStack.keys()):
-			executor.errorResolutionStack.update({str(e):[]})
+def Recover(error, obj, executor, function, *args, **kwargs):
+	logging.warning(f"Got error '{error}' from function ({function}) by {obj.name}.")
+	util.LogStack()
 
-		# ResolveError should be the only method which adds to executor.errorResolutionStack.
-		# ResolveError is itself @recoverable.
-		# So, each time we hit this point, we should also hit a corresponding ClearErrorResolutionStack() call. 
-		# If we do not, an exception is passed to the caller; if we do, the stack will be cleared upon the last resolution.
-		executor.errorRecursionDepth = executor.errorRecursionDepth + 1
+	# We have to use str(e) instead of pointers to Exception objects because multiple Exceptions will have unique addresses but will still be for the same error, as defined by string comparison.
+	if (str(error) not in executor.errorResolutionStack.keys()):
+		executor.errorResolutionStack.update({str(error):[]})
 
-		if (executor.errorRecursionDepth > len(executor.errorResolutionStack.keys())+1):
-			raise FailedErrorResolution(f"Hit infinite loop trying to resolve errors. Recursion depth: {executor.errorRecursionDepth}; STACK: {executor.errorResolutionStack}.")
+	# The executor.errorResolutionStack grows each time we invoke *this or (indirectly) executor.ResolveError().
+	# ResolveError is itself @recoverable.
+	# So, each time we hit this point, we should also hit a corresponding ClearErrorResolutionStack() call.
+	# If we do not, an exception is passed to the caller; if we do, the stack will be cleared upon the last resolution.
+	executor.errorRecursionDepth = executor.errorRecursionDepth + 1
 
-		for i, res in enumerate(executor.resolveErrorsWith):
+	if (executor.errorRecursionDepth > len(executor.errorResolutionStack.keys())+1):
+		raise FailedErrorResolution(f"Hit infinite loop trying to resolve errors. Recursion depth: {executor.errorRecursionDepth}; STACK: {executor.errorResolutionStack}.")
 
-			logging.debug(f"Checking if {res} can fix '{e}'.")
-			if (not executor.ResolveError(e, i, obj, function)): # attempt to resolve the issue; might cause us to come back here with a new error.
-				# if no resolution was attempted, there's no need to re-run the function.
-				continue
-			try:
-				logging.debug(f"Trying function ({function}) again after applying {res}.")
-				ret = function(obj, *args, **kwargs)
-				executor.ClearErrorResolutionStack(str(e)) # success!
-				logging.recovery(f"{res} successfully resolved '{e}'!")
-				logging.debug(f"Error stack is now: {executor.errorResolutionStack}")
-				return ret
-			except Exception as e2:
-				logging.debug(f"{res} failed with '{e2}'; will ignore and see if we can use another ErrorResolution to resolve '{e}'.")
+	successfullyRecovered = False
+	ret = None
+	resolvedBy = None
+	for i, res in enumerate(executor.resolveErrorsWith):
+
+		logging.debug(f"Checking if {res} can fix '{error}'.")
+		if (not executor.ResolveError(error, i, obj, function)): # attempt to resolve the issue; might cause us to come back here with a new error.
+			# if no resolution was attempted, there's no need to re-run the function.
+			continue
+		try:
+			logging.debug(f"Trying function ({function}) again after applying {res}.")
+			resolvedBy = res
+			ret = function(obj, *args, **kwargs)
+			successfullyRecovered = True
+			break
+
+		except Exception as e2:
+			if (str(error) == str(e2)):
+				logging.debug(f"{res} failed with '{e2}'; will ignore and see if we can use another ErrorResolution to resolve '{error}'.")
 				# Resolution failed. That's okay. Let's try the next.
 				# Not all ErrorResolutions will apply to all errors, so we may have to try a few before we get one that works.
-				pass
+				continue
+			else:
+				# The error changed, maybe we're making progress.
+				ret = Recover(e2, obj, executor, function, *args, **kwargs)
+				successfullyRecovered = True
+				break
 
-		#  We failed to resolve the error. Die
-		sys.tracebacklimit = 0 # traceback is NOT helpful here.
-		raise FailedErrorResolution(f"Tried and failed to resolve: {e} STACK: {executor.errorResolutionStack}. See earlier logs (in debug) for traceback.")
+	if (successfullyRecovered):
+		executor.ClearErrorResolutionStack(str(error)) # success!
+		logging.recovery(f"{resolvedBy} successfully resolved '{error}'!")
+		logging.debug(f"Error stack is now: {executor.errorResolutionStack}")
+		return ret
 
-
-# ExecutorTracker is a global singleton which keeps a record of all Executors that have been launched.
-# This can be abused quite a bit, so please try to restrict usage of this to only:
-# * Ease of use global functions
-#
-# Thanks! 
-class ExecutorTracker:
-	def __init__(this):
-		# Singletons man...
-		if "instance" not in ExecutorTracker.__dict__:
-			logging.debug(f"Creating new ExecutorTracker: {this}")
-			ExecutorTracker.instance = this
-		else:
-			return None
-
-		this.executors = [None]
-
-	@staticmethod
-	def Instance():
-		if "instance" not in ExecutorTracker.__dict__:
-			ExecutorTracker()
-		return ExecutorTracker.instance
-
-	@staticmethod
-	def Push(executor):
-		ExecutorTracker.Instance().executors.append(executor)
-
-		# Adding the executor to our list here increases its reference count.
-		# Executors are supposed to remove themselves from this list when they are deleted.
-		# A python object cannot be deleted if it has references.
-		# Thus, we forcibly decrease the reference count and rely on Exectuor's self-reporting to avoid accessing deallocated memory.
-		# This appears to cause segfaults on some systems, so we'll just live with the fact that Executors will never be destroyed.
-		# If you want your executor to stop being tracked, do it yourself. :(
-		#
-		# ctypes.pythonapi.Py_DecRef(ctypes.py_object(executor))
-
-		logging.debug(f"Now tracking Executor: {executor}")
-
-	@staticmethod
-	def Pop(executor):
-		try:
-			ExecutorTracker.Instance().executors.remove(executor)
-			logging.debug(f"No longer tracking Executor: {executor}")
-		except:
-			pass
-
-	@staticmethod
-	def GetLatest():
-		return ExecutorTracker.Instance().executors[-1]
+	#  We failed to resolve the error. Die
+	sys.tracebacklimit = 0 # traceback is NOT helpful here.
+	raise FailedErrorResolution(f"Tried and failed to resolve: {error} STACK: {executor.errorResolutionStack}. See earlier logs (in debug) for traceback.")
 
 
 # Executor: a base class for user interfaces.
@@ -1536,8 +1716,10 @@ class Executor(DataContainer, Functor):
 		this.errorRecursionDepth = 0
 		this.errorResolutionStack = {}
 		this.resolveErrorsWith = [ # order matters: FIFO (first is first).
-			"find_by_fetch",
+			'find_by_fetch',
+			'install_from_repo_with_default_package_type',
 			'install_from_repo',
+			'import_module',
 			'install_with_pip'
 		]
 
@@ -2136,165 +2318,3 @@ class Executor(DataContainer, Functor):
 
 	######## END: Fetch Locations ########
 
-
-# Use an ErrorStringParser for each "parsers" in order to avoid having to override the GetObjectFromError method and create a new class for every error you want to handle.
-# ErrorStringParsers enable ErrorResolutions to be created on a per-functionality, rather than per-error basis, reducing the total amount of duplicate code.
-# Each error has a different string. In order to get the object of the error, we have to know where the object starts and ends.
-# NOTE: this assumes only 1 object per string. Maybe fancier parsing logic can be added in the future.
-#
-# startPosition is always positive
-# endPosition is always negative
-class ErrorStringParser:
-
-	def __init__(this, applicableError, startPosition, endPosition):
-		this.applicableError = applicableError
-		this.startPosition = startPosition
-		this.endPosition = endPosition
-
-	def Parse(this, errorString):
-		end = this.endPosition
-		if (not end):
-			end = len(errorString)
-		return errorString[this.startPosition:end]
-
-
-# ErrorResolution is a Functor which can be executed when an Exception is raised.
-# The goal of this class is to do some kind of work that will fix the problem on the second try of whatever generated the error.
-class ErrorResolution(StandardFunctor):
-
-	def __init__(this, name=INVALID_NAME()):
-		super().__init__(name)
-
-		# What errors, as ErrorStringParser objects, is *this prepared to handle?
-		this.parsers = []
-
-		this.error = None
-		this.errorType = ""
-		this.errorString = ""
-		this.errorObject = ""
-		this.errorResolutionStack = {}
-
-		# Provided directly from the recoverable decorator.
-		this.optionalKWArgs["obj"] = None
-		this.optionalKWArgs["function"] = None
-
-		# We do want to know whether or not we should attempt to run whatever failed again.
-		# So, let's store that in functionSucceeded. Meaning if this.functionSucceeded, try the original method again.
-		# No rollback, by default and definitely don't throw Exceptions.
-		this.enableRollback = False
-		this.functionSucceeded = True
-		this.raiseExceptions = False
-
-		this.errorShouldBeResolved = False
-
-
-
-	# Put your logic here!
-	def Resolve(this):
-		# You get the following members:
-		# this.error (an Exception)
-		# this.errorString (a string cast of the Exception)
-		# this.errorType (a string)
-		# this.errorObjet (a string or whatever you return from GetObjectFromError())
-
-		# You get the following guarantees:
-		# *this has not been called on this particular error before.
-		# the error given is applicable to *this per this.parsers
-
-		###############################################
-		# Please throw errors if something goes wrong #
-		# Otherwise, set this.errorShouldBeResolved   #
-		###############################################
-		
-		pass
-
-
-
-	# Helper method for creating ErrorStringParsers
-	# To use this, simply take an example output and replace the object you want to extract with "OBJECT"
-	def ApplyTo(this, error, exampleString):
-		match = re.search('OBJECT', exampleString)
-		this.parsers.append(ErrorStringParser(error, match.start(), match.end() - len(exampleString)))
-
-
-	# Get the type of this.error as a string.
-	def GetErrorType(this):
-		return type(this.error).__name__
-
-
-	# Get an actionable object from the error.
-	# For example, if the error is 'ModuleNotFoundError', what is the module?
-	def GetObjectFromError(this):
-		for parser in this.parsers:
-			if (parser.applicableError != this.errorType):
-				continue
-
-			this.errorObject = parser.Parse(this.errorString)
-			return
-
-		raise ErrorResolutionError(f"{this.name} cannot parse error object from ({this.errorType}): {str(this.error)}.")
-
-
-	# Determine if this resolution method is applicable.
-	def CanProcess(this):
-		return this.GetErrorType() in [parser.applicableError for parser in this.parsers]
-
-
-	# Grab any known and necessary args from this.kwargs before any Fetch calls are made.
-	def ParseInitialArgs(this):
-		super().ParseInitialArgs()
-		if ('error' in this.kwargs):
-			this.error = this.kwargs.pop('error')
-			# Just assume the error is an actual Exception object.
-		else:
-			raise ErrorResolutionError(f"{this.name} was not given an error to resolve.")
-
-		this.errorString = str(this.error)
-		this.errorType = this.GetErrorType()
-
-		# Internal member to avoid processing duplicates
-		this.errorResolutionStack = this.executor.errorResolutionStack
-
-
-	# Error resolution is unchained.
-	def PopulateNext(this):
-		this.next = []
-
-
-	# Override of Functor method.
-	# We'll keep calling this until an error is raised.
-	def Function(this):
-		this.functionSucceeded = True
-		this.errorShouldBeResolved = True
-		
-		if (not this.CanProcess()):
-			this.errorShouldBeResolved = False
-			return this.errorResolutionStack, this.errorShouldBeResolved
-
-		if (not this.errorString in this.errorResolutionStack.keys()):
-			this.errorResolutionStack.update({this.errorString:[]})
-		
-		if (this.name in this.errorResolutionStack[this.errorString]):
-			raise FailedErrorResolution(f"{this.name} already tried and failed to resolve {this.errorType}: {this.errorString}.")
-
-		this.GetObjectFromError()
-
-		try:
-			this.Resolve()
-		except Exception as e:
-			logging.error(f"Error resolution with {this.name} failed: {e}")
-			util.LogStack()
-			this.functionSucceeded = False
-		
-		this.errorResolutionStack[this.errorString].append(this.name)
-		return this.errorResolutionStack, this.errorShouldBeResolved
-
-
-# Global Fetch() function.
-# Uses the latest registered Executor
-def Fetch(varName, default=None, fetchFrom=None, start=True, attempted=None):
-    return ExecutorTracker.GetLatest().Fetch(varName, default, fetchFrom, start, attempted)
-
-# Ease-of-use wrapper for the global Fetch()
-def f(varName, default=None, fetchFrom=None, start=True, attempted=None):
-    Fetch(varName, default, fetchFrom, start, attempted)
